(******************************************************************************)
(* generator (stack machine with condition register)                          *)
(* (C) 2010..2021 Steffen Solyga <solyga@gmx.de>                              *)
(******************************************************************************)
IMPLEMENTATION MODULE Gen;

FROM SYSTEM IMPORT ADR, BYTE;
IMPORT Lib, FIO;
FROM Scnr IMPORT TSymbol;
FROM Comp0 IMPORT TOperation, TFunction, TCmdDesc, TSet, TCmdStr;
FROM Comp0 IMPORT aNone, acMax, vcMax, pcMax, Convert;
FROM Comp IMPORT Write, GetPos, SetPos, Read;

CONST
  debug = FALSE;
  err ::= FIO.ErrorOutput;
  dbg ::= FIO.ErrorOutput;

VAR
  (* debug *)
  first: BOOLEAN;
  (* constants *)
  vals: ARRAY BOOLEAN OF TValue;

PROCEDURE DbgC( c: TCmdDesc );
VAR a: TAddress; s: TCmdStr;
BEGIN
  GetPos( a );
  Convert( c, s );
  IF first THEN FIO.WrStr( dbg, 'Gen.Dbg:  pos  cmd' ); FIO.WrLn( dbg ) END;
  first := FALSE;
  FIO.WrStr( dbg, 'Gen.Dbg: ' );
  FIO.WrCard( dbg, a, 4 );
  FIO.WrStr( dbg, ': ' );
  FIO.WrStr( dbg, s );
  FIO.WrLn( dbg );
END DbgC;

PROCEDURE DbgF( a: TAddress );
BEGIN
  FIO.WrStr( dbg, 'Gen.Dbg: Fixup ' );
  FIO.WrCard( dbg, a, 0 );
  FIO.WrLn( dbg );
END DbgF;

PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Gen: ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;

PROCEDURE ErrCard( s: ARRAY OF CHAR; x: CARDINAL );
BEGIN
  FIO.WrStr( err, 'Gen: ' );
  FIO.WrStr( err, s );
  FIO.WrChar( err, ' ' );
  FIO.WrCard( err, x, 0 );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END ErrCard;


PROCEDURE Hlt();
VAR c: TCmdDesc;
BEGIN
  c.fct := fHlt;
  IF debug THEN DbgC( c ) END;
  Write( c.fct );
END Hlt;

PROCEDURE Val( v: TValue );
VAR c: TCmdDesc;
BEGIN
  c.fct := fVal; c.val := v;
  IF debug THEN DbgC( c ) END;
  Write( c.fct ); Write( c.val );
END Val;

PROCEDURE Lod( i: TIndex );
VAR c: TCmdDesc;
BEGIN
  IF i >= vcMax THEN Err( 'Lod: Invalid index' ) END;
  c.fct := fLod; c.idx := i;
  IF debug THEN DbgC( c ) END;
  Write( c.fct ); Write( c.idx );
END Lod;

PROCEDURE Sto( i: TIndex );
VAR c: TCmdDesc;
BEGIN
  IF i >= vcMax THEN Err( 'Sto: Invalid index' ) END;
  c.fct := fSto; c.idx := i;
  IF debug THEN DbgC( c ) END;
  Write( c.fct ); Write( c.idx );
END Sto;

PROCEDURE Opr( o: TOperation );
VAR c: TCmdDesc;
BEGIN
  c.fct := fOpr; c.opr := o;
  IF debug THEN DbgC( c ) END;
  Write( c.fct ); Write( c.opr );
END Opr;

PROCEDURE Jmp( VAR a: TAddress );
VAR c: TCmdDesc;
BEGIN
  c.fct := fJmp; c.adr := a;
  IF debug THEN DbgC( c ) END;
  Write( c.fct ); GetPos( a ); Write( c.adr );
END Jmp;

PROCEDURE Jpf( VAR a: TAddress );
VAR c: TCmdDesc;
BEGIN
  c.fct := fJpf; c.adr := a;
  IF debug THEN DbgC( c ) END;
  Write( c.fct ); GetPos( a ); Write( c.adr );
END Jpf;

PROCEDURE Jpt( VAR a: TAddress );
VAR c: TCmdDesc;
BEGIN
  c.fct := fJpt; c.adr := a;
  IF debug THEN DbgC( c ) END;
  Write( c.fct ); GetPos( a ); Write( c.adr );
END Jpt;


PROCEDURE Load( VAR x: TItemDesc );
VAR a,b: TAddress;
BEGIN
  CASE x.m OF
  | mConst: Val( x.v );
  | mVar  : Lod( x.i );
  | mStack:
  | mCond : Init( a ); Init( b );
            JmpFalse( a, x ); Val( vTrue ); Jmp( b );
            Fixup( a ); Val( vFalse ); Fixup( b );
  ELSE
    ErrCard( 'Load.Bug: Invalid mode', ORD(x.m) );
  END;
  x.m := mStack;
END Load;

PROCEDURE LoadCond( VAR x: TItemDesc );
BEGIN
  Load( x ); Val( vFalse ); Opr( oNeq );
  x.m := mCond; x.n := FALSE;
END LoadCond;

PROCEDURE Wrt( x: TItemDesc );
BEGIN
  CASE x.t OF
  | tInt : Load( x ); Opr( oWrt );
  | tBool: LoadCond( x ); Opr( oWrC );  (* could be improved *)
  ELSE
    ErrCard( 'Wrt.Bug: Invalid type', ORD(x.t) );
  END;
END Wrt;

PROCEDURE Op1( s: TSymbol; VAR x: TItemDesc );
BEGIN
  IF x.m = mConst THEN
    CASE s OF
    | sMinus             : x.v := -x.v;
    | sNot               : x.v := vals[ x.v = vFalse ];
    | sAnd,sOr,sEql..sGeq:
    ELSE
      ErrCard( 'Op1.Calc.Bug: Invalid symbol', ORD(s) );
    END;
  ELSE
    CASE s OF
    | sMinus             : Load( x ); Opr( oNeg );
    | sNot               : IF x.m # mCond THEN LoadCond( x ) END; x.n := ~ x.n;
    | sAnd,sOr,sEql..sGeq: IF x.m = mCond THEN Load( x ) END;
    ELSE
      ErrCard( 'Op1.Gen.Bug: Invalid symbol', ORD(s) );
    END;
  END;
END Op1;

PROCEDURE Op2( s: TSymbol; VAR x: TItemDesc; y: TItemDesc );
VAR o: TOperation;

  PROCEDURE SetOpr();
  BEGIN
    CASE s OF
    | sPlus : o := oAdd;
    | sMinus: o := oSub;
    | sMul  : o := oMul;
    | sDiv  : o := oDiv;
    | sOr   : o := oOr;
    | sAnd  : o := oAnd;
    | sEql  : o := oEql;
    | sNeq  : o := oNeq;
    | sLss  : o := oLss;
    | sLeq  : o := oLeq;
    | sGtr  : o := oGtr;
    | sGeq  : o := oGeq;
    ELSE
      ErrCard( 'Op2.SetOpr.Bug: Invalid symbol', ORD(s) );
    END;
  END SetOpr;

  PROCEDURE Calc();
  BEGIN
    CASE o OF
    | oAdd: x.v := x.v + y.v;
    | oSub: x.v := x.v - y.v;
    | oMul: x.v := x.v  *  y.v;
    | oDiv: x.v := x.v DIV y.v;
    | oOr : x.v := TValue( TSet(x.v) + TSet(y.v) );
    | oAnd: x.v := TValue( TSet(x.v) * TSet(y.v) );
    | oEql: x.v := vals[ x.v =  y.v ];
    | oNeq: x.v := vals[ x.v #  y.v ];
    | oLss: x.v := vals[ x.v <  y.v ];
    | oLeq: x.v := vals[ x.v <= y.v ];
    | oGtr: x.v := vals[ x.v >  y.v ];
    | oGeq: x.v := vals[ x.v >= y.v ];
    ELSE
      ErrCard( 'Op2.Calc.Bug: Invalid operation', ORD(o) );
    END;
  END Calc;

  PROCEDURE InvOpr();
  BEGIN
    CASE o OF
    | oSub: o := oBus;
    | oDiv: o := oVid;
    | oEql: o := oNeq;
    | oNeq: o := oEql;
    | oLss: o := oGeq;
    | oLeq: o := oGtr;
    | oGtr: o := oLeq;
    | oGeq: o := oLss;
    ELSE
    END;
  END InvOpr;

BEGIN
  SetOpr();
  IF ( x.m = mConst ) & ( y.m = mConst ) THEN
    Calc();
  ELSE
    IF x.m = mCond THEN Err( 'Op2.Bug: Conditional mode' ) END;
    IF ( x.m # mStack ) & ( y.m = mStack ) THEN InvOpr() END;
    Load( x ); Load( y ); Opr( o );
    IF ( oEql <= o ) & ( o <= oGeq ) THEN x.m := mCond; x.n := FALSE END;
  END;
END Op2;

PROCEDURE Asg( x, y: TItemDesc );
BEGIN
  IF x.m # mVar THEN Err( 'Illegal assignment' ) END;
  Load( y ); Sto( x.i );
END Asg;

PROCEDURE Label( VAR a: TAddress );
BEGIN
  GetPos( a );
END Label;

PROCEDURE Init( VAR a: TAddress );
BEGIN
  a := aNone;
END Init;

PROCEDURE JmpFalse( VAR a: TAddress; x: TItemDesc );
BEGIN
  CASE x.m OF
  | mConst: IF x.v # vFalse THEN RETURN END; Jmp( a );
  | mCond : IF x.n THEN Jpt( a ) ELSE Jpf( a ) END;
  ELSE
    LoadCond( x ); Jpf( a );
  END;
END JmpFalse;

PROCEDURE Fixup( a: TAddress );
VAR pos, nxt: TAddress;
BEGIN
  GetPos( pos );
  WHILE a # aNone DO
    IF a > pos THEN ErrCard( 'Fixup.Bug: Invalid address', a ) END;
    IF debug THEN DbgF( a ) END;
    SetPos( a ); Read( nxt );
    SetPos( a ); Write( pos );
    a := nxt;
  END;
  SetPos( pos );
END Fixup;


PROCEDURE InitModule();
BEGIN
  first := TRUE;

  Lib.Fill( ADR(vals), SIZE(vals), 0C );
  vals[ FALSE ] := vFalse;
  vals[ TRUE  ] := vTrue;
END InitModule;

BEGIN
  InitModule();
END Gen.
